from collections.abc import Iterable, Iterator, Mapping, Sequence
from datetime import datetime
from functools import cache
from typing import Any, Optional, Union

from . import http_client
from .product import Product


def search(
    dataset_ids: Iterable[Union[str, int]] = [],
    bbox: Optional[Sequence[float]] = None,
    start_time: Union[str, datetime, None] = None,
    end_time: Union[str, datetime, None] = None,
    count: int = 100,
    params: dict[str, Any] = {},
    timeout: Optional[float] = 120,
) -> "Search":
    """Searches products on G-Portal with the given parameters.

    Note that the return value is a [`Search`][gportal.search.Search] instance.
    See its documentation for details on how to get the number of matches and iterate over results.

    Args:
        dataset_ids: List of dataset IDs.
        bbox: Bounding box of coordinates.
        start_time: Observation start time.
        end_time: Observation end time.
        count: Number of products per page.
        params: Additional search parameters. See G-Portal User's Manual.
        timeout: Timeout in seconds.

    Returns:
        A [Search][gportal.search.Search] instance that represents deferred query.
    """
    params = params.copy()

    if dataset_ids:
        params["datasetId"] = ",".join(map(str, dataset_ids))

    if bbox:
        params["bbox"] = ",".join(map(str, bbox))

    if isinstance(start_time, datetime):
        params["startTime"] = start_time.isoformat()
    elif start_time:
        params["startTime"] = start_time

    if isinstance(end_time, datetime):
        params["endTime"] = end_time.isoformat()
    elif end_time:
        params["endTime"] = end_time

    params["count"] = count

    return Search(params, timeout=timeout)


class Search:
    """Search query for G-Portal Catalogue Service API.

    For API details, see G-Portal User's Manual Appendix 7:
    https://gportal.jaxa.jp/gpr/assets/mng_upload/COMMON/upload/GPortalUserManual_en.pdf

    Attributes:
        params: Search parameters.
        timeout: Timeout in seconds.
    """

    def __init__(self, params: Mapping[str, Any], timeout: Optional[float] = None):
        self.params: Mapping[str, Any] = params
        self.timeout: Optional[float] = timeout

    def __repr__(self) -> str:
        return f"<gportal.Search params={self.params}>"

    @cache
    def matched(self) -> Optional[int]:
        """Gets the number of products matched the query.

        This method is cached, so it is safe to call it multiple times.

        Returns:
            The number of products matched the query.
        """
        response = self._request({"count": 0})
        return response.get("properties", {}).get("numberOfRecordsMatched")

    def products(self) -> Iterator[Product]:
        """Yields products matched the query.

        Yields:
            A [Product][gportal.product.Product] instance.
        """
        for page in self.pages():
            for product_dict in page.get("features", []):
                yield Product(product_dict)

    def pages(self) -> Iterator[dict[str, Any]]:
        """Iterates a search request with pagination.

        Yields:
            A dictionary of GeoJSON FeatureCollection.
        """
        start_index = 1

        while True:
            page = self._request({"startIndex": start_index})
            yield page

            properties = page["properties"]
            start_index += properties["numberOfRecordsReturned"]

            if start_index > properties["numberOfRecordsMatched"]:
                break

    def _request(self, extra_params: dict[str, Any]) -> dict[str, Any]:
        return http_client.get(
            "/csw/csw",
            params={
                **self.params,
                "service": "CSW",
                "version": "3.0.0",
                "request": "GetRecords",
                "outputFormat": "application/json",
                **extra_params,
            },
            timeout=self.timeout,
        )



from collections.abc import Mapping
from typing import Any, Optional
from urllib.parse import urljoin

import requests

BASE_URL = "https://gportal.jaxa.jp/"


def get(path: str, params: Optional[Mapping[str, Any]] = None, timeout: Optional[float] = None) -> Any:
    url = urljoin(BASE_URL, path)
    response = requests.get(url, params=params, timeout=timeout)
    response.raise_for_status()
    return response.json()



import re
from datetime import datetime
from typing import Any, Optional


class Product:
    """Wrapper of a product GeoJSON in search results.

    Attributes:
        geojson: Original dictionary that represents GeoJSON Feature.
    """

    def __init__(self, geojson: dict[str, Any]):
        self.geojson: dict[str, Any] = geojson

    def __repr__(self) -> str:
        if self.id is None:
            return super().__repr__()
        else:
            return f"<gportal.Product id={self.id}>"

    @property
    def __geo_interface__(self) -> dict[str, Any]:
        return self.geojson

    @property
    def geometry(self) -> dict[str, Any]:
        """GeoJSON geometry."""
        return self.geojson.get("geometry", {})

    @property
    def properties(self) -> dict[str, Any]:
        """GeoJSON properties."""
        return self.geojson.get("properties", {})

    @property
    def id(self) -> Optional[str]:
        """Granule ID."""
        return self.properties.get("identifier")

    @property
    def dataset_id(self) -> Optional[str]:
        """Dataset ID belonging to."""
        return self.properties.get("gpp", {}).get("datasetId")

    @property
    def start_time(self) -> datetime:
        """Observation start time."""
        isoformat = re.sub(r"Z$", "+00:00", self.properties["beginPosition"])  # For Python 3.10 or former
        return datetime.fromisoformat(isoformat)

    @property
    def end_time(self) -> datetime:
        """Observation end time."""
        isoformat = re.sub(r"Z$", "+00:00", self.properties["endPosition"])  # For Python 3.10 or former
        return datetime.fromisoformat(isoformat)

    @property
    def data_url(self) -> Optional[str]:
        """URL of the product file."""
        return self.properties.get("product", {}).get("fileName")

    @property
    def data_path(self) -> Optional[str]:
        """Path to the product file on the SFTP server."""
        if self.data_url is None:
            return None

        return self.data_url.replace("https://gportal.jaxa.jp/download/", "", 1)

    def get_browse_url(self, type: str = "browse") -> Optional[str]:
        """URL of the browse image.

        Args:
            type: Type of the image. Basically, "browse", "sub-browse" or "thumbnail".
        """
        items = self.properties.get("browse")
        if items is None:
            return None

        type = {
            "thumbnail": "thm",
            "browse": "br",
            "sub-browse": "sb",
        }.get(type, type)

        for item in items:
            if f"/img/{type}" in item["fileName"]:
                return item["fileName"]

        return None

    def flatten_properties(self) -> dict[str, Any]:
        """Flattens the nested properties.

        This method processes nested properties from the following keys:

        - `product`: Each key is prefixed with `product` (e.g. `product.fileName` => `productFileName`).
        - `browse`: Keys are the image type (e.g. browse-br, browse-thm), and their values are `fileName`.
        - `gpp`: Keys and values are directly merged.

        Returns:
            The flattened properties.
        """
        properties = self.properties.copy()

        if "product" in properties and isinstance(properties["product"], dict):
            for key, value in properties["product"].items():
                properties[f"product{key[0].upper()}{key[1:]}"] = value

            del properties["product"]

        if "browse" in properties and isinstance(properties["browse"], list):
            for browse in properties["browse"]:
                match = re.search(r"\/img\/([\w]+)\/", browse.get("fileName", ""))
                if match:
                    properties[f"browse-{match.group(1)}"] = browse["fileName"]

            del properties["browse"]

        if "gpp" in properties and isinstance(properties["gpp"], dict):
            properties.update(properties["gpp"])
            del properties["gpp"]

        return properties

    def to_dict(self) -> dict[str, Any]:
        return self.geojson

    def to_flat_properties_dict(self) -> dict[str, Any]:
        """Converts to a dictionary of GeoJSON Feature with flattened properties.

        Returns:
            A dictionary of GeoJSON Feature with flattened properties.
        """
        return {
            **self.geojson,
            "properties": self.flatten_properties(),
        }



import os.path
import re
from collections.abc import Iterable
from typing import Optional, Union

from paramiko.sftp_client import SFTPClient
from paramiko.transport import Transport

import gportal

from .product import Product


def download(
    target: Union[str, Product, Iterable[Union[str, Product]]],
    local_dir: str = ".",
    username: Optional[str] = None,
    password: Optional[str] = None,
) -> None:
    """Downloads files to a local directory via SFTP.

    Args:
        target: Remote path, Product object, or a list of them.
        local_dir: Local directory to download to.
        username: G-Portal username. If not provided, the value of `gportal.username` is used.
        password: G-Portal password. If not provided, the value of `gportal.password` is used.
    """
    with SFTP.connect(username, password) as sftp:
        sftp.download(target, local_dir)


class SFTP:
    """Wrapper of the G-Portal SFTP interface.

    Attributes:
        client: An instance of [`paramiko.SFTPClient`][paramiko.sftp_client.SFTPClient].
    """

    HOST = "ftp.gportal.jaxa.jp"
    PORT = 2051

    def __init__(self, sftp_client: SFTPClient):
        self.client: SFTPClient = sftp_client

    def __enter__(self):  # type: ignore[no-untyped-def]
        return self

    def __exit__(self, exc_type, exc_value, traceback):  # type: ignore[no-untyped-def]
        self.close()

    @classmethod
    def connect(cls, username: Optional[str] = None, password: Optional[str] = None) -> "SFTP":
        """Opens an SFTP session with G-Portal.

        Args:
            username: G-Portal username. If not provided, the value of `gportal.username` is used.
            password: G-Portal password. If not provided, the value of `gportal.password` is used.

        Returns:
            An instance of [`SFTP`][gportal.sftp.SFTP].
        """
        username = username or gportal.username
        password = password or gportal.password

        if username is None or password is None:
            raise ValueError("username and password are required")

        transport = Transport((cls.HOST, cls.PORT))
        transport.connect(username=username, password=password)

        sftp_client = transport.open_sftp_client()
        if sftp_client is None:  # will never happen
            raise RuntimeError("Failed to open SFTP session")

        return cls(sftp_client)

    def close(self) -> None:
        """Closes the SFTP session."""
        self.client.close()

    def listdir(self, path: str = "/", /, filter_pattern: Optional[str] = None, fullpath: bool = False) -> list[str]:
        """Returns a list containing the names of the entries in the given path.

        Wraps [`paramiko.SFTPClient.listdir`][paramiko.sftp_client.SFTPClient.listdir].

        Args:
            path: Remote path to list. It must be absolute.
            filter_pattern: Regular expression to filter the entries.
            fullpath: If `True`, the returned list contains full paths of the entries.

        Returns:
            A list containing the names of the entries.
        """
        self._reset_cwd()
        entries = self.client.listdir(path)

        if filter_pattern:
            entries = [entry for entry in entries if re.search(filter_pattern, entry)]

        if fullpath:
            return [os.path.join(path, entry) for entry in entries]
        else:
            return entries

    def download(self, target: Union[str, Product, Iterable[Union[str, Product]]], local_dir: str) -> None:
        """Downloads files to a local directory.

        Args:
            target: Remote path, Product object, or a list of them.
            local_dir: Local directory to download to.

        Raises:
            ValueError: If the given product has no URL to download.
        """
        self._reset_cwd()

        if isinstance(target, Iterable):
            targets = target
        else:
            targets = [target]

        for target in targets:
            if isinstance(target, Product):
                if target.data_path is None:
                    raise ValueError(f"Product {target.id} has no URL to download")

                target = target.data_path

            self.client.get(target, os.path.join(local_dir, os.path.basename(target)))

    def _reset_cwd(self) -> None:
        self.client.chdir()



__all__ = ["cached_datasets", "datasets", "search", "SFTP", "download"]

import os

from .dataset import cached_datasets, datasets
from .search import search
from .sftp import SFTP, download

username = os.getenv("GPORTAL_USERNAME", None)
password = os.getenv("GPORTAL_PASSWORD", None)



import json
import re
from functools import cache
from pathlib import Path
from typing import Any, Union

from . import http_client

Datasets = dict[str, Union["Datasets", list[str]]]


def datasets() -> Datasets:
    """Fetches the dataset tree from G-Portal.

    The tree's structure corresponds to the "spacecraft/sensor" search tree of the Web UI.
    The leaves are dataset IDs.

    Note:
        This depends on an undocumented API of G-Portal, which is subject to change.
        If this function is broken, use [`cached_datasets()`][gportal.dataset.cached_datasets] instead.

    Returns:
        The dictionary of the dataset tree, where the leaves are dataset IDs.
    """

    def build_datasets(tree: list[dict[str, Any]], root: bool = True) -> dict[str, Any]:
        datasets = {}

        for node in tree:
            if root:
                title = re.sub(r"<img[^>]*>", "", node["title"]).rstrip()
            else:
                title = node["title"]

            children = node.get("children")
            if children:
                datasets[title] = build_datasets(children, root=False)
            else:
                datasets[title] = node["dataset"].split(",")

        return datasets

    raw_tree = http_client.get("/gpr/search/service/satsensor.json")
    return build_datasets(raw_tree)


@cache
def cached_datasets() -> Datasets:
    """Loads the dataset tree from the cache included in this package.

    Returns:
        The dictionary of the dataset tree, where the leaves are dataset IDs.
    """
    dataset_path = Path(__file__).parent / "data/datasets.json"
    with dataset_path.open() as f:
        datasets: dict[str, Any] = json.load(f)
        return datasets
